@@ -13,7 +13,6 @@ import click
 from distributed import Scheduler
 from distributed.security import Security
 from distributed.utils import get_ip_interface, ignoring
-from distributed.http import HTTPScheduler
 from distributed.cli.utils import (check_python_3, install_signal_handlers,
                                    uri_from_host_port)
 from distributed.preloading import preload_modules
@@ -38,12 +37,8 @@ pem_file_option_type = click.Path(exists=True, resolve_path=True)
 @click.option('--tls-key', type=pem_file_option_type, default=None,
               help="private key file for TLS (in PEM format)")
 # XXX default port (or URI) values should be centralized somewhere
-@click.option('--http-port', type=int, default=9786,
-              help="HTTP port for JSON API diagnostics")
 @click.option('--bokeh-port', type=int, default=8787,
               help="Bokeh port for visual diagnostics")
-@click.option('--bokeh-internal-port', type=int, default=None,
-              help="Deprecated. Use --bokeh-port instead")
 @click.option('--bokeh/--no-bokeh', '_bokeh', default=True, show_default=True,
               required=False, help="Launch Bokeh Web UI")
 @click.option('--show/--no-show', default=False, help="Show web UI")
@@ -51,8 +46,6 @@ pem_file_option_type = click.Path(exists=True, resolve_path=True)
               help="IP addresses to whitelist for bokeh.")
 @click.option('--bokeh-prefix', type=str, default=None,
               help="Prefix for the bokeh app")
-@click.option('--prefix', type=str, default=None,
-              help="Deprecated, see --bokeh-prefix")
 @click.option('--use-xheaders', type=bool, default=False, show_default=True,
               help="User xheaders in bokeh app for ssl termination in header")
 @click.option('--pid-file', type=str, default='',
@@ -65,21 +58,11 @@ pem_file_option_type = click.Path(exists=True, resolve_path=True)
               help="Directory to place scheduler files")
 @click.option('--preload', type=str, multiple=True,
               help='Module that should be loaded by each worker process like "foo.bar" or "/path/to/foo.py"')
-def main(host, port, http_port, bokeh_port, bokeh_internal_port, show, _bokeh,
+def main(host, port, bokeh_port, show, _bokeh,
          bokeh_whitelist, bokeh_prefix, use_xheaders, pid_file, scheduler_file,
-         interface, local_directory, preload, prefix, tls_ca_file, tls_cert,
+         interface, local_directory, preload, tls_ca_file, tls_cert,
          tls_key):
 
-    if bokeh_internal_port:
-        print("The --bokeh-internal-port keyword has been removed.\n"
-              "The internal bokeh server is now the default bokeh server.\n"
-              "Use --bokeh-port %d instead" % bokeh_internal_port)
-        sys.exit(1)
-
-    if prefix:
-        print("The --prefix keyword has moved to --bokeh-prefix")
-        sys.exit(1)
-
     sec = Security(tls_ca_file=tls_ca_file,
                    tls_scheduler_cert=tls_cert,
                    tls_scheduler_key=tls_key,
@@ -122,7 +105,7 @@ def main(host, port, http_port, bokeh_port, bokeh_internal_port, show, _bokeh,
     loop = IOLoop.current()
     logger.info('-' * 47)
 
-    services = {('http', http_port): HTTPScheduler}
+    services = {}
     if _bokeh:
         with ignoring(ImportError):
             from distributed.bokeh.scheduler import BokehScheduler
@@ -136,6 +119,9 @@ def main(host, port, http_port, bokeh_port, bokeh_internal_port, show, _bokeh,
 
     logger.info('Local Directory: %26s', local_directory)
     logger.info('-' * 47)
+
+    install_signal_handlers(loop)
+
     try:
         loop.start()
         loop.close()
@@ -148,7 +134,6 @@ def main(host, port, http_port, bokeh_port, bokeh_internal_port, show, _bokeh,
 
 
 def go():
-    install_signal_handlers()
     check_python_3()
     main()
 
