@@ -22,14 +22,14 @@ except ImportError:
     from toolz import pluck
 from tornado.gen import Return
 from tornado import gen
-from tornado.ioloop import IOLoop, PeriodicCallback
+from tornado.ioloop import IOLoop
 from tornado.locks import Event
 
 from . import profile
 from .batched import BatchedSend
 from .comm import get_address_host, get_local_address_for
 from .comm.utils import offload
-from .config import config
+from .config import config, log_format
 from .compatibility import unicode, get_thread_identity
 from .core import (error_message, CommClosedError,
                    rpc, pingpong, coerce_to_address)
@@ -44,12 +44,15 @@ from .threadpoolexecutor import ThreadPoolExecutor, secede as tpe_secede
 from .utils import (funcname, get_ip, has_arg, _maybe_complex, log_errors,
                     ignoring, validate_key, mp_context, import_file,
                     silence_logging, thread_state, json_load_robust, key_split,
-                    format_bytes)
+                    format_bytes, DequeHandler, ThrottledGC, PeriodicCallback)
 from .utils_comm import pack_data, gather_from_workers
 
 _ncores = mp_context.cpu_count()
 
 logger = logging.getLogger(__name__)
+deque_handler = DequeHandler(n=config.get('log-length', 10000))
+deque_handler.setFormatter(logging.Formatter(log_format))
+logger.addHandler(deque_handler)
 
 LOG_PDB = config.get('pdb-on-err')
 
@@ -119,9 +122,18 @@ class WorkerBase(ServerNode):
         self.memory_limit = memory_limit
         self.paused = False
 
-        self.memory_target_fraction = config.get('worker-memory-target', 0.6)
-        self.memory_spill_fraction = config.get('worker-memory-spill', 0.7)
-        self.memory_pause_fraction = config.get('worker-memory-pause', 0.8)
+        if 'memory_target_fraction' in kwargs:
+            self.memory_target_fraction = kwargs.pop('memory_target_fraction')
+        else:
+            self.memory_target_fraction = config.get('worker-memory-target', 0.6)
+        if 'memory_spill_fraction' in kwargs:
+            self.memory_spill_fraction = kwargs.pop('memory_spill_fraction')
+        else:
+            self.memory_spill_fraction = config.get('worker-memory-spill', 0.7)
+        if 'memory_pause_fraction' in kwargs:
+            self.memory_pause_fraction = kwargs.pop('memory_pause_fraction')
+        else:
+            self.memory_pause_fraction = config.get('worker-memory-pause', 0.8)
 
         if self.memory_limit:
             try:
@@ -174,6 +186,7 @@ class WorkerBase(ServerNode):
             'call_stack': self.get_call_stack,
             'profile': self.get_profile,
             'profile_metadata': self.get_profile_metadata,
+            'get_logs': self.get_logs,
             'keys': self.keys,
         }
 
@@ -182,16 +195,16 @@ class WorkerBase(ServerNode):
                                          **kwargs)
 
         pc = PeriodicCallback(self.heartbeat,
-                              self.heartbeat_interval,
-                              io_loop=self.loop)
+                              self.heartbeat_interval)
         self.periodic_callbacks['heartbeat'] = pc
         self._address = contact_address
 
-        self._memory_monitoring = False
-        pc = PeriodicCallback(self.memory_monitor,
-                              self.memory_monitor_interval,
-                              io_loop=self.loop)
-        self.periodic_callbacks['memory'] = pc
+        if self.memory_limit:
+            self._memory_monitoring = False
+            pc = PeriodicCallback(self.memory_monitor,
+                                  self.memory_monitor_interval)
+            self.periodic_callbacks['memory'] = pc
+        self._throttled_gc = ThrottledGC(logger=logger)
 
     @property
     def worker_address(self):
@@ -263,8 +276,7 @@ class WorkerBase(ServerNode):
                         pid=os.getpid())
                 if self.death_timeout:
                     diff = self.death_timeout - (time() - start)
-                    future = gen.with_timeout(timedelta(seconds=diff), future,
-                                              io_loop=self.loop)
+                    future = gen.with_timeout(timedelta(seconds=diff), future)
                 response = yield future
                 _end = time()
                 middle = (_start + _end) / 2
@@ -381,8 +393,7 @@ class WorkerBase(ServerNode):
         with ignoring(EnvironmentError, gen.TimeoutError):
             if report:
                 yield gen.with_timeout(timedelta(seconds=timeout),
-                                       self.scheduler.unregister(address=self.contact_address),
-                                       io_loop=self.loop)
+                                       self.scheduler.unregister(address=self.contact_address))
         self.scheduler.close_rpc()
         if isinstance(self.executor, ThreadPoolExecutor):
             self.executor.shutdown(timeout=timeout)
@@ -438,7 +449,8 @@ class WorkerBase(ServerNode):
         # logger.info("%s:%d Starts job %d, %s", self.ip, self.port, i, key)
         future = self.executor.submit(function, *args, **kwargs)
         pc = PeriodicCallback(lambda: logger.debug("future state: %s - %s",
-                                                   key, future._state), 1000, io_loop=self.loop); pc.start()
+                                                   key, future._state), 1000)
+        pc.start()
         try:
             yield future
         finally:
@@ -752,6 +764,7 @@ def apply_function(function, args, kwargs, execution_state, key,
     except Exception as e:
         msg = error_message(e)
         msg['op'] = 'task-erred'
+        msg['actual-exception'] = e
     else:
         msg = {'op': 'task-finished',
                'status': 'OK',
@@ -1007,7 +1020,14 @@ class Worker(WorkerBase):
     heartbeat_interval: int
         Milliseconds between heartbeats to scheduler
     memory_limit: int
-        Number of bytes of memory that this worker should use
+        Number of bytes of memory that this worker should use.
+        Set to zero for no limit
+    memory_target_fraction: float
+        Fraction of memory to try to stay beneath
+    memory_spill_fraction: float
+        Fraction of memory at which we start spilling to disk
+    memory_pause_fraction: float
+        Fraction of memory at which we stop running new tasks
     executor: concurrent.futures.Executor
     resources: dict
         Resources that thiw worker has like ``{'GPU': 2}``
@@ -1094,9 +1114,11 @@ class Worker(WorkerBase):
             ('constrained', 'executing'): self.transition_constrained_executing,
             ('executing', 'memory'): self.transition_executing_done,
             ('executing', 'error'): self.transition_executing_done,
+            ('executing', 'rescheduled'): self.transition_executing_done,
             ('executing', 'long-running'): self.transition_executing_long_running,
             ('long-running', 'error'): self.transition_executing_done,
             ('long-running', 'memory'): self.transition_executing_done,
+            ('long-running', 'rescheduled'): self.transition_executing_done,
         }
 
         self._dep_transitions = {
@@ -1118,27 +1140,23 @@ class Worker(WorkerBase):
         WorkerBase.__init__(self, *args, **kwargs)
 
         pc = PeriodicCallback(self.trigger_profile,
-                              config.get('profile-interval', 10),
-                              io_loop=self.loop)
+                              config.get('profile-interval', 10))
         self.periodic_callbacks['profile'] = pc
 
         pc = PeriodicCallback(self.cycle_profile,
-                              profile_cycle_interval,
-                              io_loop=self.loop)
+                              profile_cycle_interval)
         pc.start()
         self.periodic_callbacks['profile-cycle'] = pc
 
         _global_workers.append(weakref.ref(self))
 
-    def __str__(self):
+    def __repr__(self):
         return "<%s: %s, %s, stored: %d, running: %d/%d, ready: %d, comm: %d, waiting: %d>" % (
             self.__class__.__name__, self.address, self.status,
             len(self.data), len(self.executing), self.ncores,
             len(self.ready), len(self.in_flight_tasks),
             len(self.waiting_for_data))
 
-    __repr__ = __str__
-
     ################
     # Update Graph #
     ################
@@ -1190,6 +1208,8 @@ class Worker(WorkerBase):
                         self.release_key(report=False, **msg)
                     elif op == 'delete-data':
                         self.delete_data(**msg)
+                    elif op == 'steal-request':
+                        self.steal_request(**msg)
                     else:
                         logger.warning("Unknown operation %s, %s", op, msg)
 
@@ -1436,7 +1456,8 @@ class Worker(WorkerBase):
                 assert key not in self.executing
                 assert key not in self.ready
 
-            del self.waiting_for_data[key]
+            self.waiting_for_data.pop(key, None)
+
             if key in self.resource_restrictions:
                 self.constrained.append(key)
                 return 'constrained'
@@ -1495,7 +1516,7 @@ class Worker(WorkerBase):
         if self.validate:
             assert all(v >= 0 for v in self.available_resources.values())
 
-    def transition_executing_done(self, key, value=no_value):
+    def transition_executing_done(self, key, value=no_value, report=True):
         try:
             if self.validate:
                 assert key in self.executing or key in self.long_running
@@ -1528,7 +1549,7 @@ class Worker(WorkerBase):
                 if key in self.dep_state:
                     self.transition_dep(key, 'memory')
 
-            if self.batched_stream:
+            if report and self.batched_stream:
                 self.send_task_state_to_scheduler(key)
             else:
                 raise CommClosedError
@@ -1895,14 +1916,22 @@ class Worker(WorkerBase):
                 pdb.set_trace()
             raise
 
+    def steal_request(self, key):
+        state = self.task_state.get(key, None)
+
+        response = {'op': 'steal-response',
+                    'key': key,
+                    'state': state}
+        self.batched_stream.send(response)
+
+        if state in ('ready', 'waiting'):
+            self.release_key(key)
+
     def release_key(self, key, cause=None, reason=None, report=True):
         try:
             if key not in self.task_state:
                 return
             state = self.task_state.pop(key)
-            if reason == 'stolen' and state in ('executing', 'long-running', 'memory'):
-                self.task_state[key] = state
-                return
             if cause:
                 self.log.append((key, 'release-key', {'cause': cause}))
             else:
@@ -2109,18 +2138,23 @@ class Worker(WorkerBase):
                     self.digests['task-duration'].add(result['stop'] -
                                                       result['start'])
             else:
-                self.exceptions[key] = result['exception']
-                self.tracebacks[key] = result['traceback']
-                logger.warning(" Compute Failed\n"
-                               "Function:  %s\n"
-                               "args:      %s\n"
-                               "kwargs:    %s\n"
-                               "Exception: %s\n",
-                               str(funcname(function))[:1000],
-                               convert_args_to_str(args2, max_len=1000),
-                               convert_kwargs_to_str(kwargs2, max_len=1000),
-                               repr(pickle.loads(result['exception'])))
-                self.transition(key, 'error')
+                if isinstance(result.pop('actual-exception'), Reschedule):
+                    self.batched_stream.send({'op': 'reschedule', 'key': key})
+                    self.transition(key, 'rescheduled', report=False)
+                    self.release_key(key, report=False)
+                else:
+                    self.exceptions[key] = result['exception']
+                    self.tracebacks[key] = result['traceback']
+                    logger.warning(" Compute Failed\n"
+                                   "Function:  %s\n"
+                                   "args:      %s\n"
+                                   "kwargs:    %s\n"
+                                   "Exception: %s\n",
+                                   str(funcname(function))[:1000],
+                                   convert_args_to_str(args2, max_len=1000),
+                                   convert_kwargs_to_str(kwargs2, max_len=1000),
+                                   repr(pickle.loads(result['exception'])))
+                    self.transition(key, 'error')
 
             logger.debug("Send compute response to scheduler: %s, %s", key,
                          result)
@@ -2160,21 +2194,24 @@ class Worker(WorkerBase):
             return
         self._memory_monitoring = True
         total = 0
+
         proc = psutil.Process()
-        frac = proc.memory_info().rss / self.memory_limit
+        memory = proc.memory_info().rss
+        frac = memory / self.memory_limit
 
         # Pause worker threads if above 80% memory use
         if self.memory_pause_fraction and frac > self.memory_pause_fraction:
+            # Try to free some memory while in paused state
+            self._throttled_gc.collect()
             if not self.paused:
-                logger.warn("Worker is at %d percent memory usage.  "
-                            "Stopping work. "
+                logger.warn("Worker is at %d%% memory usage. Pausing worker.  "
                             "Process memory: %s -- Worker memory limit: %s",
                             int(frac * 100),
                             format_bytes(proc.memory_info().rss),
                             format_bytes(self.memory_limit))
                 self.paused = True
         elif self.paused:
-            logger.warn("Worker at %d percent memory usage. Restarting work. "
+            logger.warn("Worker is at %d%% memory usage. Resuming worker. "
                         "Process memory: %s -- Worker memory limit: %s",
                         int(frac * 100),
                         format_bytes(proc.memory_info().rss),
@@ -2186,8 +2223,8 @@ class Worker(WorkerBase):
         if self.memory_spill_fraction and frac > self.memory_spill_fraction:
             target = self.memory_limit * self.memory_target_fraction
             count = 0
-
-            while proc.memory_info().rss > target:
+            need = memory - target
+            while memory > target:
                 if not self.data.fast:
                     logger.warn("Memory use is high but worker has no data "
                                 "to store to disk.  Perhaps some other process "
@@ -2197,9 +2234,17 @@ class Worker(WorkerBase):
                                 format_bytes(self.memory_limit))
                     break
                 k, v, weight = self.data.fast.evict()
+                del k, v
                 total += weight
                 count += 1
                 yield gen.moment
+                memory = proc.memory_info().rss
+                if total > need and memory > target:
+                    # Issue a GC to ensure that the evicted data is actually
+                    # freed from memory and taken into account by the monitor
+                    # before trying to evict even more data.
+                    self._throttled_gc.collect()
+                    memory = proc.memory_info().rss
             if count:
                 logger.debug("Moved %d pieces of data data and %s to disk",
                              count, format_bytes(total))
@@ -2229,9 +2274,12 @@ class Worker(WorkerBase):
         frames = sys._current_frames()
         frames = {ident: frames[ident] for ident in active_threads}
         for ident, frame in frames.items():
-            key = key_split(active_threads[ident])
-            profile.process(frame, None, self.profile_recent, stop='_concurrent_futures_thread.py')
-            profile.process(frame, None, self.profile_keys[key], stop='_concurrent_futures_thread.py')
+            if frame is not None:
+                key = key_split(active_threads[ident])
+                profile.process(frame, None, self.profile_recent,
+                                stop='_concurrent_futures_thread.py')
+                profile.process(frame, None, self.profile_keys[key],
+                                stop='_concurrent_futures_thread.py')
         stop = time()
         if self.digests is not None:
             self.digests['profile-duration'].add(stop - start)
@@ -2304,6 +2352,14 @@ class Worker(WorkerBase):
         result = {k: profile.call_stack(frame) for k, frame in frames.items()}
         return result
 
+    def get_logs(self, comm=None, n=None):
+        if n is None:
+            L = list(deque_handler.deque)
+        else:
+            L = deque_handler.deque
+            L = [L[-i] for i in range(min(n, len(L)))]
+        return [(msg.levelname, deque_handler.format(msg)) for msg in L]
+
     def validate_key_memory(self, key):
         assert key in self.data
         assert key in self.nbytes
@@ -2465,7 +2521,7 @@ class Worker(WorkerBase):
 
         if not self._client:
             from .client import Client
-            asynchronous = get_thread_identity() == self.loop._thread_ident
+            asynchronous = self.loop is IOLoop.current()
             self._client = Client(self.scheduler.address, loop=self.loop,
                                   security=self.security,
                                   set_as_default=True,
@@ -2592,3 +2648,18 @@ def secede():
     duration = time() - thread_state.start_time
     worker.loop.add_callback(worker.maybe_transition_long_running,
                              thread_state.key, compute_duration=duration)
+
+
+class Reschedule(Exception):
+    """ Reschedule this task
+
+    Raising this exception will stop the current execution of the task and ask
+    the scheduler to reschedule this task, possibly on a different machine.
+
+    This does not guarantee that the task will move onto a different machine.
+    The scheduler will proceed through its normal heuristics to determine the
+    optimal machine to accept this task.  The machine will likely change if the
+    load across the cluster has significantly changed since first scheduling
+    the task.
+    """
+    pass
